/*
 * $Id: P5Doclet.java,v 1.1.1.1 2004/07/01 07:44:47 kenta Exp $
 *
 * Copyright 2004 Kenta Cho. Some rights reserved.
 */
package jp.gr.java_conf.abagames.p5doclet;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import com.sun.javadoc.ClassDoc;
import com.sun.javadoc.Doc;
import com.sun.javadoc.ProgramElementDoc;
import com.sun.javadoc.RootDoc;
import com.sun.javadoc.Tag;
import com.sun.tools.doclets.ClassTree;
import com.sun.tools.doclets.Configuration;
import com.sun.tools.doclets.DocletAbortException;
import com.sun.tools.doclets.SourceToHTMLConverter;
import com.sun.tools.doclets.Taglet;
import com.sun.tools.doclets.standard.ClassWriter;
import com.sun.tools.doclets.standard.ConfigurationStandard;
import com.sun.tools.doclets.standard.ConstructorSubWriter;
import com.sun.tools.doclets.standard.FieldSubWriter;
import com.sun.tools.doclets.standard.MethodSubWriter;
import com.sun.tools.doclets.standard.Standard;
import com.sun.tools.doclets.standard.SubWriterHolderWriter;
import com.sun.tools.doclets.standard.tags.SeeTaglet;
import com.sun.tools.doclets.standard.tags.SimpleTaglet;

/**
 * Doclet for the Proce55ing program.
 * Creating the applet html file including links to the API document and the source.
 */
public class P5Doclet {
	public static boolean start(RootDoc root) throws IOException {
		return P5dStandard.start(root);
	}

	public static final String APPLET_TEMPLATE_OPTION = "-applettemplate";
	public static final String SOURCE_TEMPLATE_OPTION = "-sourcetemplate";
	public static final String EXPORT_OPTION = "-export";
	public static final String BCLASSES_OPTION = "-bclasses";
	public static final String APIDOC_OPTION = "-apidoc";
	public static final String NOCOMMENTITEM_OPTION = "-nocommentitem";

	public static int optionLength(String option) {
		if (option.equals(APPLET_TEMPLATE_OPTION)
			|| option.equals(SOURCE_TEMPLATE_OPTION)
			|| option.equals(EXPORT_OPTION)
			|| option.equals(BCLASSES_OPTION))
			return 2;
		else if (
			option.equals(APIDOC_OPTION)
				|| option.equals(NOCOMMENTITEM_OPTION))
			return 1;
		return 0;
	}
}

class P5dStandard extends Standard {
	public static boolean start(RootDoc root) throws IOException {
		try {
			P5dStandard standard = new P5dStandard();
			standard.configuration = new P5dConfigurationStandard();
			standard.startP5Generation(root);
			((P5dConfigurationStandard) standard.configuration).doExport();
		} catch (DocletAbortException exc) {
			return false;
		}
		return true;
	}

	public void startP5Generation(RootDoc root) {
		configuration.setOptions(root);

		String srcDir = "." + File.separator;
		String[][] options = root.options();
		configuration.destdirname =
			"." + File.separator + "doc" + File.separator;
		String prjDir = "." + File.separator;
		boolean appletTemplateOpt = false,
			sourceTemplateOpt = false,
			bclassesOpt = false;
		for (int i = 0; i < options.length; i++) {
			String[] op = options[i];
			if (op[0].equals("-sourcepath")) {
				srcDir = op[1] + File.separator;
				File srcPath = new File(op[1]);
				prjDir = srcPath.getParent() + File.separator;
				File[] list = new File(prjDir).listFiles();
				boolean hasPrj = false;
				for (int j = 0; j < list.length; j++)
					if (list[j].getName().equals(".project")) {
						hasPrj = true;
						break;
					}
				if (!hasPrj)
					prjDir = srcPath.getPath() + File.separator;
				configuration.destdirname = prjDir + "doc" + File.separator;
			} else if (op[0].equals(P5Doclet.APPLET_TEMPLATE_OPTION)) {
				((P5dConfigurationStandard) configuration).setAppletTemplate(
					op[1]);
				appletTemplateOpt = true;
			} else if (op[0].equals(P5Doclet.SOURCE_TEMPLATE_OPTION)) {
				((P5dConfigurationStandard) configuration).setSourceTemplate(
					op[1]);
				sourceTemplateOpt = true;
			} else if (op[0].equals(P5Doclet.EXPORT_OPTION)) {
				((P5dConfigurationStandard) configuration).setExportFileName(
					op[1]);
			} else if (op[0].equals(P5Doclet.BCLASSES_OPTION)) {
				((P5dConfigurationStandard) configuration).setBClassesFile(
					op[1]);
				bclassesOpt = true;
			} else if (op[0].equals(P5Doclet.APIDOC_OPTION)) {
				P5dClassWriter.createApiDoc = true;
			} else if (op[0].equals(P5Doclet.NOCOMMENTITEM_OPTION)) {
				P5dFieldSubWriter.noCommentItem = true;
				P5dMethodSubWriter.noCommentItem = true;
			}
		}
		// TODO: handle '-converttopde' options.

		if (!appletTemplateOpt)
			((P5dConfigurationStandard) configuration).setAppletTemplate(
				prjDir + P5Taglet.APPLET_TEMPLATE_DEFULT);
		if (!sourceTemplateOpt)
			((P5dConfigurationStandard) configuration).setSourceTemplate(
				prjDir + P5Taglet.SOURCE_TEMPLATE_DEFULT);
		if (!bclassesOpt)
			((P5dConfigurationStandard) configuration).setBClassesFile(
				prjDir + P5Taglet.BCLASSES_FILENAME_DEFUALT);

		if (configuration.genSrc) {
			if (configuration.destdirname.length() > 0) {
				SourceToHTMLConverter.convertRoot(
					configuration,
					root,
					configuration.destdirname
						+ File.separator
						+ Configuration.SOURCE_OUTPUT_DIR_NAME);
			} else {
				SourceToHTMLConverter.convertRoot(
					configuration,
					root,
					Configuration.SOURCE_OUTPUT_DIR_NAME);
			}
		}
		ClassTree classtree = new ClassTree(configuration, true);
		ClassDoc[] arr = configuration.classDocCatalog.ordinaryClasses("");
		Arrays.sort(arr);

		String topClassName = arr[0].name();
		((P5dConfigurationStandard) configuration).setTopClass(
			topClassName,
			srcDir + topClassName + ".java",
			srcDir);

		// Search the classdir from '.', srcDir, '../bin', '../classes' and '../class'. 
		String clsDir = prjDir;
		if (checkFileExist(prjDir, topClassName + ".class")) {
		} else if (checkFileExist(srcDir, topClassName + ".class")) {
			clsDir = srcDir;
		} else if (
			checkFileExist(
				prjDir + File.separator + "bin",
				topClassName + ".class")) {
			clsDir = prjDir + "bin" + File.separator;
		} else if (
			checkFileExist(
				prjDir + File.separator + "classes",
				topClassName + ".class")) {
			clsDir = prjDir + "classes" + File.separator;
		} else if (
			checkFileExist(
				prjDir + File.separator + "class",
				topClassName + ".class")) {
			clsDir = prjDir + "class" + File.separator;
		}
		((P5dConfigurationStandard) configuration).classdirname = clsDir;

		for (int i = 0; i < arr.length; i++) {
			if (!(configuration.isGeneratedDoc(arr[i])
				&& arr[i].isIncluded())) {
				continue;
			}
			ClassDoc prev = (i == 0) ? null : arr[i - 1];
			ClassDoc curr = arr[i];
			ClassDoc next = (i + 1 == arr.length) ? null : arr[i + 1];
			P5dClassWriter.generate(configuration, curr, prev, next, classtree);
		}
	}

	private boolean checkFileExist(String dir, String file) {
		File tf = new File(dir + File.separator + file);
		if (tf.exists() && tf.isFile())
			return true;
		return false;
	}
}

class P5dConfigurationStandard extends ConfigurationStandard {
	private P5Taglet p5taglet;
	public String classdirname = "";

	public void setOptions(RootDoc root) {
		super.setOptions(root);
		p5taglet = new P5Taglet();
		p5taglet.setMessage(message);
		tagletManager.addCustomTag(p5taglet);
		classuse = false;
		createindex = false;
		createoverview = false;
		createtree = false;
	}

	public void setTopClass(String name, String path, String dir) {
		p5taglet.setTopClass(name, path, dir);
	}

	public void setAppletTemplate(String at) {
		p5taglet.setAppletTemplate(at);
	}

	public void setSourceTemplate(String st) {
		p5taglet.setSourceTemplate(st);
	}

	public void setExportFileName(String ef) {
		p5taglet.setExportFileName(ef);
	}

	public void setBClassesFile(String ef) {
		p5taglet.setBClassesFile(ef);
	}

	public void doExport() {
		if (destdirname.length() > 0)
			p5taglet.setDestDir(destdirname);
		if (classdirname.length() > 0)
			p5taglet.setClassDir(classdirname);
		p5taglet.doExport();
	}
}

class P5dClassWriter extends ClassWriter {
	public static boolean createApiDoc = false;

	public P5dClassWriter(
		ConfigurationStandard configuration,
		String path,
		String fileName,
		ClassDoc classDoc,
		ClassDoc prev,
		ClassDoc next,
		ClassTree classTree)
		throws IOException {
		super(configuration, path, fileName, classDoc, prev, next, classTree);
		constrSubWriter = new P5dConstructorSubWriter(this, classDoc);
		fieldSubWriter = new P5dFieldSubWriter(this, classDoc);
		methodSubWriter = new P5dMethodSubWriter(this, classDoc, configuration);
	}

	public static void generate(
		ConfigurationStandard configuration,
		ClassDoc classDoc,
		ClassDoc prev,
		ClassDoc next,
		ClassTree classTree) {
		ClassWriter clsGen;
		String fileName = classDoc.name() + ".html";
		try {
			clsGen =
				new P5dClassWriter(
					configuration,
					"",
					fileName,
					classDoc,
					prev,
					next,
					classTree);
			clsGen.generateClassFile();
			clsGen.close();
			if (!createApiDoc)
				new File(configuration.destdirname + fileName).delete();
		} catch (IOException exc) {
			configuration.standardmessage.error(
				"doclet.exception_encountered",
				exc.toString(),
				fileName);
			throw new DocletAbortException();
		}
	}

	public void generateClassFile() {
		/*String cltype =
			getText(
				classdoc.isInterface() ? "doclet.Interface" : "doclet.Class");*/
		String clname = classdoc.name();
		String label = clname; //cltype + " " + clname;

		printHtmlHeader(
			clname,
			configuration.metakeywords.getMetaKeywords(classdoc));
		navLinks(true);
		hr();
		println("<!-- ======== START OF CLASS DATA ======== -->");
		h2();
		print(label);
		h2End();
		hr();
		if (!configuration.nocomment) {
			if (classdoc.inlineTags().length > 0) {
				printInlineComment(classdoc);
				p();
			}
			generateTagInfo(classdoc);
			hr();
			p();
		} else {
			hr();
		}
		printAllMembers();
		println("<!-- ========= END OF CLASS DATA ========= -->");
		hr();
		navLinks(false);
		printBottom();
		printBodyHtmlEnd();
	}

	protected void navLinks(boolean header) {
		println("");
		if (!configuration.nonavbar) {
			if (header) {
				println("\n<!-- ========= START OF TOP NAVBAR ======= -->");
				anchor("navbar_top");
				print("\n");
				print(
					getHyperLink(
						"",
						"skip-navbar_top",
						"",
						false,
						"",
						getText("doclet.Skip_navigation_links"),
						""));
			} else {
				println("\n<!-- ======= START OF BOTTOM NAVBAR ====== -->");
				anchor("navbar_bottom");
				print("\n");
				print(
					getHyperLink(
						"",
						"skip-navbar_bottom",
						"",
						false,
						"",
						getText("doclet.Skip_navigation_links"),
						""));
			}
			table(0, "100%", 1, 0);
			tr();
			tdColspanBgcolorStyle(3, "#EEEEFF", "NavBarCell1");
			println("");
			if (header) {
				anchor("navbar_top_firstrow");
			} else {
				anchor("navbar_bottom_firstrow");
			}
			table(0, 0, 3);
			print("  ");
			trAlignVAlign("center", "top");
			navLinkClass();
			print("  ");
			trEnd();
			tableEnd();
			tdEnd();
			tdAlignVAlignRowspan("right", "top", 3);
			printUserHeaderFooter(header);
			tdEnd();
			trEnd();
			println("");
			tr();
			tdBgcolorStyle("white", "NavBarCell2");
			font("-2");
			space();
			navLinkPrevious();
			space();
			println("");
			space();
			navLinkNext();
			fontEnd();
			tdEnd();
			trEnd();
			tableEnd();
			if (header) {
				aName("skip-navbar_top");
				aEnd();
				println("\n<!-- ========= END OF TOP NAVBAR ========= -->");
			} else {
				aName("skip-navbar_bottom");
				aEnd();
				println("\n<!-- ======== END OF BOTTOM NAVBAR ======= -->");
			}
			println("");
		}
	}

	public void generateTagInfo(Doc doc) {
		configuration.tagletManager.checkTags(doc, doc.tags(), false);
		configuration.tagletManager.checkTags(doc, doc.inlineTags(), true);
		Taglet[] definedTags;
		definedTags = configuration.tagletManager.getTypeCustomTags();
		Tag[][] tagsInDoc = new Tag[definedTags.length][];
		boolean found = false;
		for (int i = 0; i < definedTags.length; i++) {
			tagsInDoc[i] = doc.tags(definedTags[i].getName());
			if (tagsInDoc[i].length > 0) {
				found = true;
			}
		}
		if (found || (doc.isClass() && ((ClassDoc) doc).isSerializable())) {
			dl();
			String output;
			for (int i = 0; i < definedTags.length; i++) {
				if (definedTags[i] instanceof SimpleTaglet) {
					output =
						((SimpleTaglet) definedTags[i]).toString(
							tagsInDoc[i],
							this);
				} else if (definedTags[i] instanceof SeeTaglet) {
					// To hide the link to the serialized form.
					output = null;
				} else {
					output = definedTags[i].toString(tagsInDoc[i]);
				}
				if (output != null && output.length() > 0) {
					configuration.tagletManager.seenCustomTag(
						definedTags[i].getName());
					print(output);
				}
			}
			dlEnd();
		}
	}

	protected void printAllMembers() {
		if (!configuration.nocomment) {
			if (prev != null) {
				println("<!-- ======== CONSTRUCTOR SUMMARY ======== -->");
				println();
				constrSubWriter.printMembersSummary();
				println();
			}
			println("<!-- =========== FIELD SUMMARY =========== -->");
			println();
			fieldSubWriter.printMembersSummary();
			println();
			println("<!-- ========== METHOD SUMMARY =========== -->");
			println();
			methodSubWriter.printMembersSummary();
			p();
		}
		if (prev != null) {
			println("<!-- ======== CONSTRUCTOR DETAIL ======== -->");
			println();
			constrSubWriter.printMembers();
			println();
		}
		//println("<!-- ============ FIELD DETAIL =========== -->");
		//println();
		//fieldSubWriter.printMembers();
		//println();
		println("<!-- ============ METHOD DETAIL ========== -->");
		println();
		methodSubWriter.printMembers();
	}
}

class P5dConstructorSubWriter extends ConstructorSubWriter {
	public P5dConstructorSubWriter(
		SubWriterHolderWriter writer,
		ClassDoc classDoc) {
		super(writer, classDoc);
	}

	protected void printSummaryType(ProgramElementDoc member) {
		// To hide the modifier.
	}
}

class P5dFieldSubWriter extends FieldSubWriter {
	public static boolean noCommentItem = false;

	public P5dFieldSubWriter(SubWriterHolderWriter writer, ClassDoc classDoc) {
		super(writer, classDoc);
	}

	protected void printModifier(ProgramElementDoc member) {
		// To hide the modifier.
	}

	protected void printSummaryLink(ClassDoc cd, ProgramElementDoc member) {
		// To remove the link in the field summary.
		String name = member.name();
		writer.bold();
		writer.print(name);
		writer.boldEnd();
	}

	public void printMembersSummary() {
		List members = new ArrayList(members(classdoc));
		if (members.size() > 0) {
			printedSummaryHeader = true;
			printSummaryHeader(classdoc);
			Collections.sort(members);
			for (int i = 0; i < members.size(); ++i) {
				ProgramElementDoc elmDoc = (ProgramElementDoc) (members.get(i));
				if (noCommentItem
					|| (elmDoc != null
						&& elmDoc.getRawCommentText() != null
						&& elmDoc.getRawCommentText().length() > 0))
					printSummaryMember(classdoc, elmDoc);
			}
			printSummaryFooter(classdoc);
		}
	}
}

class P5dMethodSubWriter extends MethodSubWriter {
	public static boolean noCommentItem = false;

	public P5dMethodSubWriter(
		SubWriterHolderWriter writer,
		ClassDoc classDoc,
		ConfigurationStandard configuration) {
		super(writer, classDoc, configuration);
	}

	protected void printModifier(ProgramElementDoc member) {
		// To hide the modifier.
	}

	public void printMembersSummary() {
		List members = new ArrayList(members(classdoc));
		if (members.size() > 0) {
			printedSummaryHeader = true;
			printSummaryHeader(classdoc);
			Collections.sort(members);
			for (int i = 0; i < members.size(); ++i) {
				ProgramElementDoc elmDoc = (ProgramElementDoc) (members.get(i));
				if (noCommentItem
					|| (elmDoc != null
						&& elmDoc.getRawCommentText() != null
						&& elmDoc.getRawCommentText().length() > 0))
					printSummaryMember(classdoc, elmDoc);
			}
			printSummaryFooter(classdoc);
		}
	}

	public void printMembers() {
		List members = members(classdoc);
		if (members.size() > 0) {
			printHeader(classdoc);
			boolean firstElm = true;
			for (int i = 0; i < members.size(); ++i) {
				ProgramElementDoc elmDoc = (ProgramElementDoc) (members.get(i));
				if (noCommentItem
					|| (elmDoc != null
						&& elmDoc.getRawCommentText() != null
						&& elmDoc.getRawCommentText().length() > 0)) {
					if (!firstElm)
						writer.printMemberHeader();
					else
						firstElm = false;
					writer.println("");
					printMember(elmDoc);
					writer.printMemberFooter();
				}
			}
			printBodyHtmlEnd(classdoc);
		}
	}

}
