/*
 * $Id: P5Taglet.java,v 1.1.1.1 2004/07/01 07:44:47 kenta Exp $
 *
 * Copyright 2004 Kenta Cho. Some rights reserved.
 */
package jp.gr.java_conf.abagames.p5doclet;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.channels.FileChannel;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import com.sun.javadoc.Tag;
import com.sun.tools.doclets.MessageRetriever;
import com.sun.tools.doclets.Taglet;

/**
 * Taglet works with the P5Doclet.
 * Retrieve @p5 tag and create html files and the jar file.
 */
public class P5Taglet implements Taglet {
	private static String NAME = "p5";

	public String getName() {
		return NAME;
	}

	public boolean inConstructor() {
		return true;
	}

	public boolean inField() {
		return true;
	}

	public boolean inMethod() {
		return true;
	}

	public boolean inOverview() {
		return true;
	}

	public boolean inPackage() {
		return true;
	}

	public boolean inType() {
		return true;
	}

	public boolean isInlineTag() {
		return false;
	}

	public static void register(Map tagletMap) {
		P5Taglet tag = new P5Taglet();
		Taglet t = (Taglet) tagletMap.get(tag.getName());
		if (t != null) {
			tagletMap.remove(tag.getName());
		}
		tagletMap.put(tag.getName(), tag);
	}

	public String toString(Tag tag) {
		return "";
	}

	private MessageRetriever message;

	public void setMessage(MessageRetriever m) {
		message = m;
	}

	private static final String TITLE_KEYWORD = "title";
	private static final String DESCRIPTION_KEYWORD = "description";
	private static final String SIZE_KEYWORD = "size";

	private String appletTemplate, sourceTemplate;
	private String exportFileName;
	private String bclassesFile;
	private String title;
	private StringBuffer description;
	private final String DEFAULT_SIZE = "200";
	private String width, height;

	public static String APPLET_TEMPLATE_DEFULT = "applet_template.html";
	public static String SOURCE_TEMPLATE_DEFULT = "source_template.html";
	public static String EXPORT_FILENAME_DEFAULT = "index.html";
	public static String BCLASSES_FILENAME_DEFUALT = "bclasses.jar";
	
	public P5Taglet() {
		// Store the default values.
		appletTemplate = APPLET_TEMPLATE_DEFULT;
		sourceTemplate = SOURCE_TEMPLATE_DEFULT;
		exportFileName = EXPORT_FILENAME_DEFAULT;
		bclassesFile = BCLASSES_FILENAME_DEFUALT;
		width = height = DEFAULT_SIZE;
		description = new StringBuffer();
	}
	
	public void setAppletTemplate(String at) {
		appletTemplate = at;
	}

	public void setSourceTemplate(String st) {
		sourceTemplate = st;
	}

	public void setExportFileName(String ef) {
		exportFileName = ef;
	}

	public void setBClassesFile(String bf) {
		bclassesFile = bf;
	}

	public String toString(Tag[] tag) {
		StringBuffer rsl = new StringBuffer();
		rsl.append("<p>");
		for (int i = 0; i < tag.length; i++) {
			StringTokenizer tkn = new StringTokenizer(tag[i].text());
			if (!tkn.hasMoreTokens())
				continue;
			String type = tkn.nextToken();
			if (type.equals(TITLE_KEYWORD)) {
				if (!tkn.hasMoreTokens())
					continue;
				title = tkn.nextToken();
			} else if (type.equals(DESCRIPTION_KEYWORD)) {
				String dsc = tag[i].text();
				dsc = dsc.substring(dsc.indexOf(DESCRIPTION_KEYWORD) + DESCRIPTION_KEYWORD.length());
				StringTokenizer ltkn = new StringTokenizer(dsc, "\n");				
				handleDescription(ltkn, rsl);
			} else if (type.equals(SIZE_KEYWORD)) {
				if (!tkn.hasMoreTokens())
					continue;
				width = tkn.nextToken();
				if (!tkn.hasMoreTokens())
					continue;
				height = tkn.nextToken();
			}
		}
		rsl.append("</p>");
		return rsl.toString();
	}

	private void handleDescription(StringTokenizer tkn, StringBuffer rsl) {
		StringBuffer r = new StringBuffer();
		rsl.append("<b>Description:</b><br>");
		boolean itemized = false;
		while (tkn.hasMoreTokens()) {
			String s = tkn.nextToken();
			s = s.trim();
			if (s.startsWith("-")) {
				if (!itemized) {
					itemized = true;
					r.append("<ul>");
				}
				r.append("<li>");
				s = s.substring(1);
				s = s.trim();
				r.append(s);
			} else {
				if (itemized) {
					r.append("</ul>");
					itemized = false;
				}
				r.append(s);
				r.append("<br>");
			}
		}
		if (itemized)
			r.append("</ul>");
		description.append(r);
		description.append("<br>");
		rsl.append(r);
	}

	private String topClassName, topClassPath, topClassDir;

	public void setTopClass(String name, String path, String dir) {
		topClassName = name;
		topClassPath = path;
		topClassDir = dir;
		title = topClassName + " : Built with Processing";
		// Parse the 'size( , )' to get the value of width and height.
		try {
			String src = readFile(topClassPath);
			Pattern p =
				Pattern.compile("size\\s*\\(\\s*[0-9]+\\s*,\\s*[0-9]+\\s*\\)");
			Matcher m = p.matcher(src);
			m.find();
			String sz = m.group();
			Pattern pn = Pattern.compile("[0-9]+");
			Matcher mn = pn.matcher(sz);
			mn.find();
			width = mn.group();
			mn.find();
			height = mn.group();
		} catch (Exception e) {
			System.out.println("Can't parse the 'size', using default.");
			width = height = DEFAULT_SIZE;
		}
	}

	private String destDir = "." + File.separator;
	
	public void setDestDir(String dir) {
		destDir = dir;
	}

	private String classDir = "." + File.separator;

	public void setClassDir(String dir) {
		classDir = dir;
	}

	private String src, srcFileName;

	public void doExport() {
		// Export to the web.
		Runtime runtime = Runtime.getRuntime();
		try {
			Process prc;
			// TODO: when both 'class' and 'jar' don't find, try to use 'javac' to create class.
			//prc = runtime.exec("javac " + pathToClass);
			//prc.waitFor();
			// Copy the original source file to destDir.
			if (checkFileExist(topClassDir, topClassName + ".pde" )) {
				srcFileName = topClassName + ".pde";
			} else {
				srcFileName = topClassName + ".java";
			}
			copyFile(new File(topClassDir + srcFileName), new File(destDir + srcFileName));
			System.out.println("copy: " + topClassDir + srcFileName + " to " + destDir + srcFileName);

			if (checkFileExist(classDir, topClassName + ".jar")) { 
				// If 'jar' exists in the class directory, use it and don't create the another.
				copyFile(new File(topClassName + ".jar"), new File(destDir + topClassName + ".jar"));
				System.out.println("copy: " + topClassName + ".jar to " + destDir + topClassName + ".jar");
			} else {
				// Create the jar file to export as the applet.
				copyFile(new File(bclassesFile), new File(destDir + topClassName + ".jar"));
				System.out.println("copy: " + bclassesFile + " to " + destDir + topClassName + ".jar");
				String cmd = "jar ufM " + destDir + topClassName + ".jar *.class";
				prc = runtime.exec(cmd, null, new File(classDir));
				System.out.println("exec@" + classDir + ": " + cmd);
				prc.waitFor();
				printErrorStream(prc.getErrorStream());
				// Add the data files.
				File dataDir = new File(classDir + "data" + File.separator);
				if (dataDir.exists() && dataDir.isDirectory()) {
					cmd = "jar ufM " + destDir + topClassName + ".jar .";
					prc = runtime.exec(cmd, null, dataDir);
					System.out.println("exec@" + dataDir.getAbsolutePath() + ": " + cmd);
					prc.waitFor();
					printErrorStream(prc.getErrorStream());
				}
			}

			String srcHtml = readFile(sourceTemplate);
			src = readFile(topClassDir + srcFileName);
			if (Pattern.compile("\\[\\[escapesrc\\]\\]").matcher(srcHtml).find()) {
				// Escape letters for showing the source code in a pre element.
				src = src.replaceAll("&", "&amp;");
				src = src.replaceAll("<", "&lt;");
				src = src.replaceAll(">", "&gt;");
				srcHtml = srcHtml.replaceAll("\\[\\[escapesrc\\]\\]", "");
			}
			try {
				Pattern p =
					Pattern.compile("\\[\\[replacetab\\s*:\\s*[0-9]+\\s*\\]\\]");
				Matcher m = p.matcher(srcHtml);
				m.find();
				String sz = m.group();
				Pattern pn = Pattern.compile("[0-9]+");
				Matcher mn = pn.matcher(sz);
				mn.find();
				int spcLgt = Integer.parseInt(mn.group());
				StringBuffer spc =  new StringBuffer();
				for (int i = 0 ; i < spcLgt; i++)
					spc.append(" ");
				src = src.replaceAll("\\t", spc.toString());
				srcHtml = srcHtml.replaceAll("\\[\\[replacetab\\s*:\\s*[0-9]+\\s*\\]\\]", "");
			} catch (Exception e) {}
			srcHtml = replaceTagsInTemplate(srcHtml);
			writeFile(
				topClassName + "_src.html",
				srcHtml);

			String appletHtml = readFile(appletTemplate);
			appletHtml = replaceTagsInTemplate(appletHtml);
			writeFile(exportFileName, appletHtml);
		} catch (Exception e) {
			if (message != null)
				message.error(e.getMessage());
		}
	}

	private String replaceTagsInTemplate(String html) {
		String rsl = html;
		rsl = rsl.replaceAll("\\[\\[title\\]\\]", title);
		rsl = rsl.replaceAll("\\[\\[description\\]\\]", description.toString());
		rsl = rsl.replaceAll("\\[\\[width\\]\\]", width);
		rsl = rsl.replaceAll("\\[\\[height\\]\\]", height);
		rsl = rsl.replaceAll("\\[\\[name\\]\\]", topClassName);
		rsl = rsl.replaceAll("\\[\\[jarfile\\]\\]", topClassName + ".jar");
		rsl = rsl.replaceAll("\\[\\[sourcefile\\]\\]", srcFileName);
		rsl = rsl.replaceAll("\\[\\[classfile\\]\\]", topClassName + ".class");
		rsl = rsl.replaceAll("\\[\\[sourcehtml\\]\\]", topClassName + "_src.html");
		rsl = rsl.replaceAll("\\[\\[javadochtml\\]\\]", topClassName + ".html");
		rsl = rsl.replaceAll("\\[\\[source\\]\\]", src);
		return rsl;
	}

	private String readFile(String fileName) throws IOException {
		File file = new File(fileName);
		BufferedReader reader = new BufferedReader(new FileReader(file));
		StringBuffer rsl = new StringBuffer();
		for (;;) {
			String line = reader.readLine();
			if (line == null)
				break;
			rsl.append(line);
			rsl.append("\n");
		}
		reader.close();
		return rsl.toString();
	}

	private void writeFile(String fileName, String content)
		throws IOException {
		File file = new File(destDir + fileName);
		BufferedWriter writer = new BufferedWriter(new FileWriter(file));
		writer.write(content);
		writer.close();
	}

	public void copyFile(File in, File out) throws IOException {
		FileChannel sourceChannel = new FileInputStream(in).getChannel();
		FileChannel destinationChannel = new FileOutputStream(out).getChannel();
		sourceChannel.transferTo(0, sourceChannel.size(), destinationChannel);
		sourceChannel.close();
		destinationChannel.close();
	}
	
	private boolean checkFileExist(String dir, String file) {
		File tf = new File(dir + File.separator + file);
		if (tf.exists() && tf.isFile())
			return true;
		return false;
	}

	private void printErrorStream(InputStream stderr) {
		BufferedReader br = new BufferedReader(new InputStreamReader(stderr));
		String line = null;
		try {
			while ( (line = br.readLine()) != null)
				System.out.println(line);
			br.close();
		} catch (IOException e) {}
	}
}
